#include <malloc.h>
#include <stdio.h>
#include "allocate.h"
#include "debug.h"

allocator allNew(size_t siz, int cAlloc, char *sz) {
  allocator all = malloc(sizeof(struct _allocator));

  /* Round cAlloc up to nearest power of 2, for efficiency in pAllocated() */
  cAlloc--;
  all->iLogCAlloc = 0;
  while (cAlloc > 0) {
    cAlloc >>= 1;
    all->iLogCAlloc++;
  }
  cAlloc = (1 << all->iLogCAlloc);

  all->siz = siz;
  all->cAlloc = cAlloc;
  all->cRemaining = 0;
  /* pNext can stay uninitialized */
  all->cBlocks = 0;
  all->stChunks = stNew(4);
  all->sz = sz;

  return all;
}

void FreeAll(allocator all) {  /* frees this allocator and all the memory allocated by it */
  while (iStackEntries(all->stChunks) > 0)  
    free(pPopStack(all->stChunks));
  FreeStack(all->stChunks);
  free(all);
}

void ClearAll(allocator all) {   /* keeps the allocator, but frees all the memory it's allocated so far; can be reused with same creation parameters */
  while (iStackEntries(all->stChunks) > 0)  
    free(pPopStack(all->stChunks));
  FreeStack(all->stChunks);
  all->stChunks = stNew(4);
  all->cRemaining = 0;
  all->cBlocks = 0;
}

void *pAllocate(allocator all) {
  return pAllocateMany(all, 1);
}

/* Allocate a contiguous group of blocks. */

void *pAllocateMany(allocator all, int cBlocks) { 
  void *p;

  if (all->cRemaining < cBlocks) {   /* not enough space in the current chunk, allocate more */

    if (all->sz) 
      DebugPrintf(dbg_allocate, " [%d %s allocated] ", iAllocatedSoFar(all), all->sz);

    if (all->cAlloc - cBlocks  < all->cRemaining) {
      /* If the caller has requested more blocks than fit in a single
	 chunk, we have to do an exceptional malloc() rather than
	 using part of a standard chunk.  Actually the condition we use for this
	 is a bit more general, and will apply to some other cases where the
	 caller has requested more than half a chunk: We do a malloc()
	 here if allocating a new chunk and taking the request from
	 there would leave less space left in the new chunk than in
	 the current one.  In that case, it's better to allocate the
	 request separately and keep the leftover space in the current
	 chunk. */
      p = malloc(all->siz * cBlocks);
      assert (p != NULL);
      PushStack(all->stChunks, p);
      all->cBlocks += cBlocks;
      return p;                   /* Note: we RETURN HERE in this special case */
    }
    
    /* Just throw away the all->cRemaining blocks (wastefully)
       and create a new chunk of blocks. */

    all->pNext = malloc(all->siz * all->cAlloc);
    assert(all->pNext != NULL);
    PushStack(all->stChunks, all->pNext);
    all->cBlocks += all->cAlloc;
    all->cRemaining = all->cAlloc;
  }

  /* Carve caller's request off the current chunk. */

  p = all->pNext;
  all->pNext = (void*)((char*)all->pNext + cBlocks * all->siz);
  all->cRemaining -= cBlocks;
  assert(all->cRemaining >= 0);
  return p;
}

void *pAllocated(allocator all, int i) {    /* returns the ith thing that was allocated (like an array index) */
  assert(i >= 0);
  assert(i < iAllocatedSoFar(all));
  return all->stChunks->ap[i >> all->iLogCAlloc] + all->siz * (i & (all->cAlloc-1));  /* implement div and mod with bit operations (cAlloc is a power of 2) */
}
  
int iAllocatedSoFar(allocator all) {
  return all->cBlocks - all->cRemaining;
}

void *pAllocatedCopyall(allocator all) {           
  void *p = malloc(all->siz * all->cBlocks);   /* all the space */
  void *q = p;                                 /* next free space */
  size_t sizChunk = (all->siz << all->iLogCAlloc);
  int i;
  for (i=0; i<iStackEntries(all->stChunks)-1; i++) {  /* copy all chunks except last */
    memcpy(q, all->stChunks->ap[i], sizChunk);
    q += sizChunk;
  }
  memcpy(q, all->stChunks->ap[i], all->siz * (all->cAlloc - all->cRemaining));  /* copy last */
  return p;
}


/* We will iterate BACKWARDS over the allocated blocks.  This allows
   a slightly simpler implementation, plus it restricts to the
   blocks that were already allocated at the time we called
   allitNew().  As usual, iterator returns NULL when it runs out. */

alliter allitNew(allocator all) {
  alliter allit = malloc(sizeof(struct _alliter));
  allit->all = all;
  allit->iChunk = iStackEntries(all->stChunks) - 1;
  allit->pPrev = all->pNext;   /* notionally, we just returned the first object that hasn't yet been allocated (might be first in its chunk) */
  allit->pFirst = all->stChunks->ap[allit->iChunk];
  return allit;
}

void FreeAllit(alliter allit) {
  free(allit);
}

void *pNextAllocated(alliter allit) {
  if (allit->pPrev == allit->pFirst) {  /* must move to previous chunk */
    if (--allit->iChunk < 0)  
      return NULL;
    allit->pFirst = allit->all->stChunks->ap[allit->iChunk];                      /* start of new chunk */ 
    allit->pPrev =  allit->pFirst + (allit->all->siz << allit->all->iLogCAlloc);  /* put pPrev just past end of new chunk */
  }
  return (allit->pPrev -= allit->all->siz);
}
