#ifndef _ALLOCATE_H
#define _ALLOCATE_H

#include <malloc.h>
#include "stack.h"

/* !!! Could write another type that has very low space overhead in
   the case where only 0 or a few things are allocated.  (E.g., for
   w->allFa.)  It should just store a size and a pointer to a linked
   list of allocated objects (each of which also has room for an extra
   link pointer).  This means that the overhead is one pointer per
   item, plus one size overall.  This list is only allowed to get up
   to length 10 or so (checked every time something new is allocated).
   If it gets beyond that, then the size should be reset to 0 (as a
   flag), and the pointer should be reset to point to a newly
   allocated structure that contains both the pointer to that linked
   list and also a struct _allocator that is used for remaining items,
   whose cAlloc is perhaps 16 (or could instead do a repeated-doubling
   version of the allocator, which would be just as efficient except
   in pAllocated(), where it would be less so). */

typedef struct _allocator {  /* all: simple memory allocator that 
				allocates memory in chunks and can only free all of it at once.
				Caller can carve individual object-size blocks, or an array
				of such blocks (e.g., a string), off the end of the current 
				chunk.  If there's not enough space left in the current chunk, 
				the rest of it is thrown away (wastefully) and we allocate 
				a new chunk. 
                                We can iterate over all the blocks allocated so far. 
                                We can return a copy of all the blocks aggregated together.
                                Random access to a block (by number) is possible in linear
                                time, but it's rather slower than for stacks.

				   (Maybe it makes sense to combine stacks and allocators. !!!

				      Start with allocators; implement popping by freeing chunks;
				      maybe let size of chunk i be c*a^i for a=1 OR a=2.
				      The random access is a little slower than for stacks, because
				      it takes two array accesses not one.  On the other hand,
				      elements once allocated don't move, as they do for stacks.
				      Only the chunks that point to them might move (when the chunk
				      array is reallocated).  We could even allow freeing
				      of objects in the middle: the first time this is attempted,
				      the allocator creates a second, internal allocator that lists
				      all free slots; the internal allocator can use pure stack discipline,
				      so there is no recursion necessary here.
				      In principle, we could have each chunk maintain its
				      count of full slots, so if everything in the chunk happens
				      to be freed we can deallocate the chunk.  
				      A caller must not free an object twice: this could be
				      integrity-checked by a routine that writes 1 into the
				      first byte of every free slot, and then writes 0 into 
				      the first byte of every free slot, crashing if 0 has
				      already been written there.) */

  size_t siz;          /* size of block to return */
  int cAlloc;          /* number of blocks in an allocated chunk.  Must be a power of 2. */
  int cRemaining;      /* number of blocks left in the current chunk */
  short iLogCAlloc;    /* log2(cAlloc) */
  void *pNext;         /* next available block. */
  int cBlocks;         /* total blocks ALLOCATED so far (not returned) - useful for keeping statistics */
  stack stChunks;      /* all the chunks.  Used when we free the allocator, and for iteration. */
  char *sz;            /* plural name of block type - for messages when we allocate more - NULL if we don't want such messages*/
} *allocator;

typedef struct _alliter {
  allocator all;
  int iChunk;   /* number of current chunk */
  void *pPrev;  /* block we just returned */
  void *pFirst; /* first block in this chunk - if we just returned it, time to move back to previous chunk */
} *alliter;


allocator allNew(size_t siz, int cAlloc, char *sz);
void FreeAll(allocator all);
void ClearAll(allocator all);
void *pAllocate(allocator all);
void *pAllocateMany(allocator all, int cBlocks);
int iAllocatedSoFar(allocator all);
void *pAllocated(allocator all, int i);
void *pAllocatedCopyall(allocator all);

alliter allitNew(allocator all);
void FreeAllit(alliter allit);
void *pNextAllocated(alliter allit);

#endif  /* _ALLOCATE_H */
