#define DEBUG_DEFINE_VARS   /* makes debug.h define variables, not just declare them */
#include "debug.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <time.h>
#include <signal.h>
#include "hash.h"
#include "stack.h"

#define STACK_WARNING_SIZE 20

char *szDebugElapsedTime(void);


/* -------------------------------------------------- */

/* symbolic names passed with debug calls are turned into numbers here */

int iFromName(char *sz) {
  static hash hsh = NULL;   

  if (hsh==NULL) 
    hsh = hshNew(100, "symbolic names for debug routines");

  return iHashedFromSz(hsh, sz, NULL);
}

/* -------------------------------------------------- */

/* A version of printf with some extra bells and whistles.

   Returns immediately unless the flag f is true.

   Always prints to stderr.  

   An initial or final string of spaces, tabs, OR newlines in the
   format string is interpreted as a REQUEST for a minimum amount of
   whitespace to insulate the output from non-whitespace.  It may not
   all be printed.  For example, if we try to print " foobar", but the
   last call to DebugPrintf ended in two spaces already, then we only
   need to print one more space before "foobar."  If the last call
   ended in five spaces or a newline, then we don't print any space
   (note that newline trumps any amount of horizontal space).

   %| in the string is ignored, but may be used initially or finally
   to prevent a string of whitespace from being initial or final.  This
   is useful if the caller is about to print, or has just printed, something
   on stderr without using DebugPrintf.

   %: within the string prints a timestamp, if dbg_time is true.

   Support for indentation.  "%5>" within the string says to indent 5
   additional spaces from now on, starting at the next newline.  
   "%>" is equivalent to "%3>".  To indent using an arbitrary prefix 
   (i.e., not necessarily a list of spaces), use "%0>prefix\n".  (If
   \n is omitted, the new prefix will go to the end of the string.)
   All these are canceled by "%<", which may appear in a later call
   to DebugPrintf.


     !!! it would be nice to allow "%0>[%s/%d]\n", or perhaps even "%%d>" or %0>%:\n"
         We will currently do a very poor job of handling these if somebody tries them!
	 (Note that %% is ambiguous above ...)

     !!! if string starts with "%>", and we have newlines pending, should
         probably start by adding the new prefix? ...

     !!! could do a much nicer job of spacing around the timestamp.
	 We want foo%:bar to become foo [time] bar or foobar according to dbg_time.
	 But we want foo %: bar to become foo [time] bar or foo bar.
	 And we want \n%:bar to become \n[time] bar or \nbar.  
	 That is, the spacing is the same as if we were printing "foo ", then " [time] " or "",
	 then " bar" using three separate calls to DebugPrintf.  

	 Should actually rewrite this properly: use a little subroutine that
	 copies one character from szFormat to szFormatNew, and keeps track of
	 the last character copied and the length of its run.  Give it a flag
	 whereby it can decline to copy if the existing run is too long.
	 That will simplify and clarify a lot of the below, at some cost to
	 efficiency but who cares.

     !!! maybe have %: interact with indentation, so that elapsed time
         is measured on the current level of indentation ... ?

     !!! multiple uses of %: in the same call are currently identical; but this
         is confusing behavior.
*/

#define DEBUG_DEFAULT_INDENT 3
#define DEBUG_TIMESTAMP_LEN_MAX 100

void DebugPrintf(bool f, char *szFormat, ...) {

  static stack stszPrefixes = NULL;   /* indentation prefixes */
  static int ichPrefixes = 0;         /* total # chars for current indentation prefixes */
  static int iSpaces = 0;
  static int iTabs = 0;
  static int iNewlines = 10000;       /* Initialized to reflect the fact that we start
				       the program on a new line, and don't need any
				       initial blank lines. */
  char *pchNewFormatMax, *pchPrefixTerminatingNewline=P_DEBUG_GARBAGE, *szNewFormat, *szElapsedTime=NULL;
  va_list ap;

  if (!f) return;

  if (stszPrefixes == NULL) 
    stszPrefixes = stNew(10);

  /* chop off initial whitespace as necessary. */
  {
    int i;
    if ((i=strspn(szFormat, " ")) > 0) {  /* leading spaces */
      if (iSpaces >= i || iNewlines > 0)
	szFormat+=i;
      else 
	szFormat += iSpaces;
    } else if ((i=strspn(szFormat, "\t")) > 0) {  /* leading tabs */
      if (iTabs >= i || iNewlines > 0)
	szFormat+=i;
      else 
	szFormat += iSpaces;
    } else if ((i=strspn(szFormat, "\n")) > 0) { /* leading newlines */
      if (iNewlines >= i)
	szFormat += i;
      else
	szFormat += iNewlines;
    }
  }

  /* allocate szNewFormat that is guaranteed to be big enough - sloppy 
     counting, so probably too big, but that's ok */
  { 
    char *pch;
    int ich = 0;
    int iPrefixLenMax = ichPrefixes;
    for (pch = szFormat; *pch; pch++) {
      ich++;
      if (*pch == '\n')
	ich += iPrefixLenMax;
      else if (0==strncmp("%:",pch,2)) {
	if (dbg_time) {   /* otherwise we'll ignore */
	  if (szElapsedTime==NULL) szElapsedTime = szDebugElapsedTime();
	  ich += strlen(szElapsedTime);
	}
      }
      else if (0==strncmp("%0>",pch,3))   /* safe */
	iPrefixLenMax += strcspn(pch+3, "\n");   /* number of non-\n characters following %0> */
      else if (0==strncmp("%>",pch,2))   
	iPrefixLenMax += DEBUG_DEFAULT_INDENT;
      else if (*pch == '%') {
	int i = 0;
	char *pchT = pch;
	while (isdigit((int)*++pchT)) {
	  i *= 10;
	  i += (*pchT) - '0';
	}
	iPrefixLenMax += i;
      }
    }
    szNewFormat = (char *)malloc(ich+1);
    pchNewFormatMax = szNewFormat + ich;
  }


  /* fill in szNewFormat */
  {
    char *pch; 
    char *pchDst = szNewFormat;
    
    for (pch = szFormat; *pch; pch++) {
      if (*pch == '\n') {
	int i;
	*pchDst++ = *pch;
	for (i=0; i < iStackEntries(stszPrefixes); i++) {   /* follow \n with all the prefixes */
	  strcpy(pchDst, (char *)stszPrefixes->ap[i]);
	  pchDst += strlen((char *)stszPrefixes->ap[i]);
	}
      }	  
      
      else if (0==strncmp("%:",pch,2)) {
	if (dbg_time) {
	  assert(szElapsedTime);   /* should have set this above */
	  strcpy(pchDst, szElapsedTime);
	  pchDst += strlen(szElapsedTime);
	}
	pch++;    /* move pch forward to the colon */
      }

      else if (0==strncmp("%0>",pch,3)) {
	int i = strcspn(pch+3, "\n");
	char *sz = (char*)malloc(i+1);
	strncpy(sz, pch+3, i);
	sz[i] = '\0';
	PushStack(stszPrefixes, sz);
	ichPrefixes += i;
	pch += i+3;   /* move pch to the terminating \n character */
	if (*pch=='\0') pch--;  /* if it's missing, because we're at the end of the string, step it back one */
	else pchPrefixTerminatingNewline = pch;     /* remember where the \n was */
      }

      else if (*pch=='%') {
	int i = 0;
	char *pchT = pch;
	while (isdigit((int)*++pchT)) {   /* read numeric argument following % (currently don't allow decimal points) */
	  i *= 10;
	  i += (*pchT) - '0';
	}
	if (*pchT=='>') {
	  char *sz;
	  if (pchT == pch+1)     /* we just have "%>" */
	    i = DEBUG_DEFAULT_INDENT;
	  sz = (char*)malloc(i+1);
	  memset(sz, ' ', i);
	  sz[i] = '\0';
	  PushStack(stszPrefixes, sz);
	  ichPrefixes += i;
	  pch = pchT;   /* move pch to the > character */  
	} else if (*pchT=='<') {
	  char *sz;
	  assert(pchT == pch+1);   /* shouldn't have a numeric argument to "%<" */
	  assert(iStackEntries(stszPrefixes) > 0);
	  sz = pPopStack(stszPrefixes);
	  ichPrefixes -= strlen(sz);
	  free(sz);
	  pch = pchT;   /* move pch to the < character */
	} else if (*pchT=='|') {
	  pch = pchT;   /* move pch to the | character */
	} else {
	  *pchDst++ = *pch;   /* a normal % for printf; just copy it */
	}
      }

      else
	*pchDst++ = *pch;   /* copy the character */
    }
    *pchDst = '\0';
    assert(pchDst <= pchNewFormatMax);  /* make sure we allocated enuf space */
  }
	  
  /* print out message using szNewFormat and the remaining arguments */
  va_start(ap, );   
  vfprintf(stderr, szNewFormat, ap);
  va_end(ap);

  /* adjust iNewlines, iSpaces, iTabs to take final whitespace into account */
  {
    char *pch = szFormat + strlen(szFormat) - 1;   /* last char */
    char *pchT = pch;

    if (pch >= szFormat) {    /* szFormat is non-empty  (otherwise, do nothing) */

      while (--pchT >= szFormat && *pchT == *pch)     /* move pchT back before string of characters identical to last char */
	;
      switch(*pch) {
      case ' ': 
	iNewlines = iTabs = 0;
	if (pchT >= szFormat) iSpaces = 0;     /* only reset space count if we had some non-spaces in szFormat */
	iSpaces += pch - pchT;
	break;
      case '\t': 
	iNewlines = iSpaces = 0;
	if (pchT >= szFormat) iTabs = 0;     
	iTabs += pch - pchT;
	break;
      case '\n': 
	iSpaces = iTabs = 0;
	if (pchT >= szFormat) iNewlines = 0;   
	if (pch!=pchPrefixTerminatingNewline)  /* if the final \n was just terminating a prefix for %0>, it doesn't count */
	  iNewlines += pch - pchT;
	break;
      default:
	iSpaces = iTabs = iNewlines = 0;
      }
    }
  }    
      
  free(szNewFormat);
}


/* Prints elapsed time.
   Partly bypasses DebugPrintf, so that next call to DebugPrintf believes
   we're still at the start of a line -- the timestamp "doesn't take up space"
   on the new line. */

void DebugPrintTime(bool f) {

  DebugPrintf(f, "\n");
  if (f) fprintf(stderr, szDebugElapsedTime());
}
  


/* String containing time elapsed since last call and total time since first call. 
   This is in temporary storage, and will be overwritten on next call. */

#define CCH_ELAPSED_TIME_MAX 100    /* should always be enough; note that ctime is 26 characters including \0 */

char *szDebugElapsedTime(void) {
  static time_t tFirst = -1;
  static time_t tLast;
  static char sz[CCH_ELAPSED_TIME_MAX];
  sz[CCH_ELAPSED_TIME_MAX-1] = '\0';

  if (tFirst==-1) {
    tFirst = tLast = time(NULL);
    sprintf(sz, "[start 0:00:00] %s", ctime(&tLast));   /* ends in \n (change?) */
  } else {
    time_t t = time(NULL);
    int dt = t - tFirst;
    sprintf(sz, "[+%lds to %d:%d%d:%d%d] ", t - tLast, dt/10/6/10/6, dt/10/6/10%6, dt/10/6%10, dt/10%6, dt%10);
    tLast = t;
  }

  assert(sz[CCH_ELAPSED_TIME_MAX-1] == '\0');   /* if not, we overwrote our buffer */
  return sz;
}




/* -------------------------------------------------- */


typedef struct _dbgdot {    /* dot: used to print progress dots */
  int *piFreq;                  /* how often should we print the dots?  (0 for never) */
  char *szDot;                 /* usually ".", but can be any string */
  int iTotal;                  /* how many dots have we printed so far? */
} *dbgdot;

static stack stdot = NULL;


/* Register or re-register a type of dot. */

void DebugStartDots(int *piFreq, char *szDotType, char *szDot) {  
  int i = iFromName(szDotType);
  dbgdot dot = malloc(sizeof(struct _dbgdot));

  dot->piFreq = piFreq;
  dot->szDot = szDot;
  dot->iTotal = 0;
  
  if (stdot==NULL)
    stdot = stNew(4);   /* number of dot types we can initially accomodate */
  
  if (i < iStackEntries(stdot)) {   /* there is already a dot type by this name */
    free(stdot->ap[i]);                /* free the existing record */
    stdot->ap[i] = dot;                /* and replace with the new one */
  } else {
    PushStack(stdot, dot);
  }

  DebugPrintf(*dot->piFreq, " (\"%s\"=%d %s) ", szDot, *dot->piFreq, szDotType);
}

/* Count an event of a registered type, possibly printing a dot.  
   Note that this is expensive - it requires a hash even if szDotType
   is short.  So best if szDotType is short.

   (!!! Could probably avoid the hash by using string macros ... ?
   Or have DebugStartDot() return a dot for the caller to provide on
   future calls instead of a name?)
*/

void DebugDot(char *szDotType) {    
  int i = iFromName(szDotType);
  dbgdot dot;
  assert(i < iStackEntries(stdot));
  dot = stdot->ap[i];

  if (*dot->piFreq > 0 && ++(dot->iTotal) % *dot->piFreq == 0)      /* note that if piFreq is 0, we don't even count this call.*/
    DebugPrintf(1, dot->szDot);
}
  
      
/* -------------------------------------------------- */

/* check stack discipline - useful in making sure that malloc and free
   match up, for instance.

   sz is an arbitrary name, used to identify a stack of pointers.
   Calls to DebugPush and DebugPop are required to nest properly -
   only the pointer atop the stack may be popped.  Otherwise we get an
   assertion failure.

   To assist in checking for memory leaks, we print a warning whenever
   the stack size first passes STACK_WARNING_SIZE, and again whenever it 
   passes twice the size of the last warning. */

static stack stst = NULL;   /* stack of stacks */
static stack sti = NULL;    /* parallel stack of sizes at which we will next print warning */


void DebugPushBounded(char *sz, void *p, int iMax) {   /* iMax < 0 disables bounding */
  int i = iFromName(sz);
  stack st;

  if (stst==NULL) {
    stst = stNew(4);   /* number of stacks we count on initially */
    sti = stNew(4);    /* warning sizes for those stacks */
  }

  if (i < iStackEntries(stst)) 
    st = (stack)stst->ap[i];
  else {
    PushStack(stst, st = stNew(10));                   /* initial allocation for stack */
    PushStack(sti, (void *)(int)STACK_WARNING_SIZE);   /* initial warning size (hackish cast!) */
  }

  if (iMax >= 0)
    assert(iStackEntries(st) < iMax);    /* assert that there's room for another */

  PushStack(st, p);

  if (iStackEntries(st) >= (int)sti->ap[i]) {
    fprintf(stderr, " [Warning: Debug stack %s reached large size %d for first time] ", sz, (int)sti->ap[i]);
    sti->ap[i] = (void *)(2 * (int)sti->ap[i]);   /* double warning size (hackish cast!) */
  }
}

void DebugPush(char *sz, void *p) {
  DebugPushBounded(sz, p, -1);
}

void DebugPop(char *sz, void *p) {
  int i = iFromName(sz);
  stack st;

  /* find the relevant stack - it had better exist! */
  assert (stst != NULL);            
  assert (i < iStackEntries(stst));   
  st = stst->ap[i];

  assert(p==pPopStack(st));
}

/* -------------------------------------------------- */

void DebugCheckout(char *sz, int i) {   /* checkout a resource of which only i exist */
  DebugPushBounded(sz, NULL, i);
}

void DebugCheckin(char *sz) {
  DebugPop(sz, NULL);
}

/* -------------------------------------------------- */

/* assert macro is redefined to grab this.
   It sends SIGTRAP instead of using exit(1); thus, debugger
   can examine state and even resume execution.  
   Also, it starts with \n, which is nice. */

void DebugAssert(const char *szAssertion, const char *szFile, int iLine) {
  fprintf (stderr, "\n%s:%u: failed assertion `%s'\n", szFile, iLine, szAssertion);
  sigsend(P_PID, P_MYID, SIGTRAP);   
}

/* -------------------------------------------------- */

    
