#ifndef _DEBUG_H
#define _DEBUG_H

#include "bool.h"

/* lots of good things to do here ...
    a nice class that handles printing of progress dots.
    ability to turn different pieces of tracing (and assertion?) on and off by keyword.
    indentation of tracing,
    handling of newlines ...
    */

#ifdef DEBUG_DEFINE_VARS   /* true only when we're included into debug.c */
#define DEBUG_VAR(type, var, value) type var = value
#else
#define DEBUG_VAR(type, var, value) extern type var
#endif /* DEBUG_DEFINE_VARS */

#define DEBUG_FLAG(flag, value) DEBUG_VAR(int, flag, value)


/* -------------- program-specific stuff --------------- */


/* variables to switch various pieces of tracing code on and off;
   can be changed by the program or in the debugger. */

DEBUG_FLAG(dbg_time, 1);        /* allow DebugPrintf to print timestamps for %: */
DEBUG_FLAG(dbg_allocate, 1);    /* new blocks of identical objects allocated */
DEBUG_FLAG(dbg_resize, 1);      /* data structures grown or shrunk (e.g., hash tables) */
	   
/* -------------- general debugging stuff --------------- */

/* garbage values - use this to initialize things that aren't supposed
   to need initialization, to make sure that we break if we do try to
   use the value.  Probably a good idea to try a few other values here
   (e.g., very positive or very tiny) to make sure that behavior
   doesn't change. */

#define I_DEBUG_GARBAGE -12345             /* implausible and recognizable */
#define P_DEBUG_GARBAGE (void *)0x123      /* will crash if dereferenced, on most systems */
#define R_DEBUG_GARBAGE -1.2345e300;       /* would be better to use a signaling NaN value */


void DebugPush(char *sz, void *p);
void DebugPushBounded(char *sz, void *p, int iMax);
void DebugPop(char *sz, void *p);

void DebugCheckout(char *sz, int i);
void DebugCheckin(char *sz);


void DebugPrintf(bool f, char *szFormat, ...);
void DebugPrintTime(bool f);

void DebugStartDots(int *piFreq, char *szDotType, char *szDot);
void DebugDot(char *szDotType);

void DebugAssert(const char *, const char *, int);


/* cribbed from assert.h, but modified to use DebugAssert rather than __assert.
   I assume ANSI standard, i.e., __STDC__ should be defined. 
   Note that standard assert.h should NOT be included as well, as it undefs assert. 

   I've added "expect" -- a version that is sure to evaluate the expression,
   even if NDEBUG is set.  This is dandy for checking return values, etc.
   I generally leave NDEBUG on, but the distinction ensures safety and is also 
   helpful when reading code. 

   I've also added "assertw" (weak assert), which is functionally
   identical to assert but is pragmatically different.  It should be
   used for assertions that signal that the program is working as we
   think, but whose failure probably wouldn't break anything. */

#ifdef	NDEBUG
#define	assert(EX) ((void)0)
#define expect(EX) ((void)(EX))
#else
#define	assert(EX) (void)((EX) || (DebugAssert(#EX, __FILE__, __LINE__), 0))
#define expect(EX) (void)((EX) || (DebugAssert(#EX, __FILE__, __LINE__), 0))
#endif	/* NDEBUG */

#define assertw(EX) assert(EX)

#endif /*  _DEBUG_H */



#if _TESTCODE

  #include "debug.h"

  DebugPrintf(1, "This is test #%d.\n", 5);
  DebugPrintf(1, " [testing] ");
  DebugPrintf(1, " ... ");
  DebugPrintf(1, "  Ok, from now on %>all new lines will be indented.  ");
  DebugPrintf(1, " You'll see it soon:\nline 1\nline 2\n");
  DebugPrintf(1, "\nline 3\n");
  DebugPrintf(1, "%12>\n12 more spaces of indentation!");
  DebugPrintf(1, "%0>(foo)\nAbout to enter the foo routine:");
  DebugPrintf(1, "\nline 1");
  DebugPrintf(1, " ... ");
  DebugPrintf(1, "\nline 2");
  DebugPrintf(1, " ... ");
  DebugPrintf(1, "%<\nExited the foo routine");
  DebugPrintf(1, "%<\nUnindent\nContinue%<\nExiting");

#endif /* _TESTCODE */

