/* !!! I use type tag "i" in 3 ways -- return values of iHash, iHash % whatever, and 
   iHashedFromSz.  In particular the third is conceptually different and should have
   its own type tag.  Note that the second is used as an array index internal to hsh,
   while the third might be used as an array index external to hsh. */

#include <stdlib.h>
#include "hash.h"  
#include "string.h"
#include "allocate.h"
#include "prime.h"
#include "debug.h"
#include "bool.h"  

#define I_COLLISION_INCREMENT 123456791   /* arbitrary large prime number -- will be relatively prime to any smaller hashtable size (note: can check primality with Unix command "factor").
					     !!! since we're picking prime hashtable sizes anyway, faster to make the increment a power of 2 ... ? */
#define SZ_ALLOC_CHUNK 2000               /* allocate memory for strings in chunks of this many bytes */
#define iHash iHashPJWRandXor16           /* which hash function to use */
#define MAX_STR_LEN 200                   /* longest string (not counting \0) we're willing to hash if we're using randomized hashing */
#define HASH_FULL_FRAC 0.7                /* when hash table gets this full, it's time to resize */

void ResizeHshMaybe(hash hsh);
unsigned int _iHashedFromSz(hash hsh, char *sz);
char *strdup_nofree(char *sz, allocator all);        
unsigned int iHashPJW(const char* x);
unsigned int iHashIspell (char *p);
unsigned long iHashElf(const unsigned char *name);
unsigned int iHashPJWRandMult(const unsigned char *x);
unsigned int iHashPJWRandMult16(const unsigned char *x);
unsigned int iHashPJWRandXor16(const unsigned char *x);


unsigned int iHshEntries(hash hsh) {   
  return hsh->csz; 
}


/* Make a new empty hash, with initial capacity at least csz.
   If szName is non-NULL, resize messages will be printed 
   (if the appropriate dbg_ flags are set). */

hash hshNew(unsigned int csz, char *szName) {   
  hash hsh;
  hsh = (hash)malloc(sizeof(*hsh));
  
  /* round size up to a prime, at least 17 */
  if (csz < 16) csz = 16;
  csz = uliNearbyPrime(csz);
    
  /* allocate arrays */    
  hsh->ai = (int *)malloc(csz * sizeof(int));
  hsh->asz = (char **)calloc(csz, sizeof(char *));   /* init to NULL */
  assert(hsh->asz[0] == NULL);                       /* portability check: calloc() actually just initializes to 0,
							and ANSI C allows NULL to be represented as other than 0.
							If we're running on the rare machine where this assertion fails,
							then we should replace calloc() here and in ResizeHshMaybe()
							with a malloc plus explicit initialization of elements. */

  /* initialize other fields */
  hsh->allsz = allNew(sizeof(char), SZ_ALLOC_CHUNK, NULL /*"hashchars"*/);   
  hsh->csz = 0;
  hsh->cszMac = csz;
  hsh->diCollision = I_COLLISION_INCREMENT % hsh->cszMac;  /* avoid doing a % operation at time of collision */
  hsh->iNext = 0;
  hsh->iLookups = 0;
  hsh->iCollisions = 0;
  hsh->szName = szName ? strdup_nofree(szName, hsh->allsz) : NULL;   

  return hsh;
}


/* free the hash table */
void FreeHsh(hash hsh) {
  FreeAll(hsh->allsz);
  free(hsh->asz);
  free(hsh->ai);
  free(hsh);
}


/* Returns a pointer to the hash value for the given key, so that the
   caller can change it.  If the key wasn't already in the table, adds
   it with the value iDefault.  This bypasses the whole business of
   assigning sequential integers.  

   Shares code with iHashedFromSz(), which could do its work through
   us if we supported its extra psz argument.
 */

unsigned int *piHashedFromSz(hash hsh, char *sz, int iDefault) {
  unsigned int i = _iHashedFromSz(hsh, sz);

  if (hsh->asz[i]==NULL) {                          /* we didn't find key */
    hsh->asz[i] = strdup_nofree(sz, hsh->allsz);    /* add copy of key to hash */
    hsh->ai[i] = iDefault;                          /* assign integer to it */
    hsh->csz++;
  }

  return &hsh->ai[i];
}



/* Find integer corresponding to sz in the hash table hsh.  If sz
   doesn't yet appear in hsh, put it in and assign the next integer to it. 
   Also, if psz is non-NULL, set *psz to point to a canonical copy of sz,
   which the hash table stores in any case. 
   Shares code with piHashedFromSz(). */

unsigned int iHashedFromSz(hash hsh, char *sz, char **psz) {

  unsigned int i = _iHashedFromSz(hsh, sz);

  if (hsh->asz[i]==NULL) {                          /* we didn't find key */
    hsh->asz[i] = strdup_nofree(sz, hsh->allsz);    /* add copy of key to hash */
    hsh->ai[i] = hsh->iNext++;                      /* assign next integer to it */
    hsh->csz++;
  }

  if (psz) *psz = hsh->asz[i];

  return hsh->ai[i];
}  


/* Like iHashedFromSz(), but if sz doesn't appear in hsh yet,
   just returns -1 instead of adding it. */

int iHashedFromSzNoCreate(hash hsh, char *sz) {
  unsigned int i = _iHashedFromSz(hsh, sz);

  if (hsh->asz[i]==NULL) return -1;   /* we didn't find key */
  else return hsh->ai[i];
}


  
  

unsigned int _iHashedFromSz(hash hsh, char *sz) {   
  
  unsigned int i;

  ResizeHshMaybe(hsh);
  i = iHash(sz) % (hsh->cszMac);   /* hash the string and turn it into array index */

  ++hsh->iLookups;
  while (hsh->asz[i] && strcmp(sz, hsh->asz[i])) {   /* so long as hash entry is present and wrong */
    ++hsh->iCollisions;
    /* if (hsh->iLookups < 100) fprintf(stderr, "%p Collision%c %d %d -> %d: strlens %d %d: occupancy %d/%d\t%s\t%s\n", hsh, iHash(sz)==iHash(hsh->asz[i])?'!':' ', iHash(sz), iHash(hsh->asz[i]), i, strlen(sz), strlen(hsh->asz[i]), hsh->csz, hsh->cszMac, sz, hsh->asz[i]);  */
    i += hsh->diCollision;   
    if (i >= hsh->cszMac) i -= hsh->cszMac;  /* avoids a % operation */
  }

  return i;
}



unsigned int iHshUpcoming(hash hsh) {   /* next number that will be assigned; if 
					   caller doesn't like it, caller can skip
					   it using SkipHsh() */
  return hsh->iNext;
}


void SkipHsh(hash hsh) {        /* see iHshUpcoming() */
  hsh->iNext++;
}

/* ------------------------------ */


/* resize the hash table if it's getting too full.  
   We roughly double the size.  Various tests for "too full" are possible here.  !!!

   !!! If I ever add a "delete from table" method, I should add code to shrink the
   table if it's getting too empty. */ 

void ResizeHshMaybe(hash hsh) {

  int j;

  if (hsh->csz > HASH_FULL_FRAC * hsh->cszMac) {     /* !!! Another possible test: if (hsh->csz == hsh->cszMac || (hsh->iLookups >= 100 && hsh->iCollisions > 2*hsh->iLookups))  */

    /* save old arrays*/
    int cszMacOld = hsh->cszMac;
    char **aszOld = hsh->asz;   
    int *aiOld = hsh->ai;
    
    /* replace with new arrays. */
    if (hsh->szName) 
      DebugPrintf(dbg_resize, "\n[Hashtable \"%s\" at %d/%d (%d%%) occupancy, %d lookups, %.2f probes/lookup; ", hsh->szName, hsh->csz, hsh->cszMac, (100*hsh->csz)/hsh->cszMac, hsh->iLookups, 1 + ((float)hsh->iCollisions)/hsh->iLookups);
    hsh->cszMac = uliNearbyPrime(hsh->cszMac * 2);  
    hsh->asz = (char **)calloc(hsh->cszMac, sizeof(char *));     /* init to NULL */
    assert(hsh->asz[0] == NULL);                       /* extra portability check -- see note at calloc() above. */
    hsh->ai = (unsigned int *)malloc(hsh->cszMac * sizeof(unsigned int));
    hsh->diCollision = I_COLLISION_INCREMENT % hsh->cszMac;  /* avoid doing a % operation at time of collision */
    if (hsh->szName)
      DebugPrintf(dbg_resize, "increasing to %d --", hsh->cszMac);

    /* transfer contents */
    for (j=0; j<cszMacOld; j++) {
      if (aszOld[j]) {
	int i = iHash(aszOld[j]) % (hsh->cszMac);
	while (hsh->asz[i]) {   /* while there's something (else!) in the desired cell */
	  i += hsh->diCollision;   
	  if (i >= hsh->cszMac) i -= hsh->cszMac;  /* avoids a % operation */
	}
	hsh->asz[i] = aszOld[j];
	hsh->ai[i] = aiOld[j];
      }
    }
    if (hsh->szName)
      DebugPrintf(dbg_resize, " done] ");

    /* free old arrays, reset statistics */
    free(aszOld);
    free(aiOld);
    hsh->iLookups = 0;
    hsh->iCollisions = 0;
  }
}


/* -------------------------------------------------- */

hashp hshpNew(unsigned int csz, char *szName) {
  hashp hshp;
  hshp = (hashp)malloc(sizeof(*hshp));

  hshp->hsh = hshNew(csz, szName);
  hshp->st = stNew(csz);
  
  return hshp;
}
  
void FreeHshp(hashp hshp) {
  FreeHsh(hshp->hsh);
  FreeStack(hshp->st);
  free(hshp);
}

unsigned int iHshpEntries(hashp hshp) {
  return iHshEntries(hshp->hsh);
}

/* like iHashedFromSz(), except that it takes care of storing a copy of
   the print name away if it's new. */

unsigned int iHashedPrintableFromSz(hashp hshp, char *sz) {
  char *szCopy;
  unsigned int i;

  assert(iHshpUpcoming(hshp) == iStackEntries(hshp->st));

  i = iHashedFromSz(hshp->hsh, sz, &szCopy);
  if (i == iStackEntries(hshp->st))    /* we've added a new name (greater # than previously seen), so store its name away */
    PushStack(hshp->st, szCopy);

  return i;
}

/* Version that returns -1 (rather than adding sz) if sz is not already
   in the table.  Like iHashedFromSzNoCreate. */

int iHashedPrintableFromSzNoCreate(hashp hshp, char *sz) {
  return iHashedFromSzNoCreate(hshp->hsh, sz);
}

  
char *szHshpPrintname(hashp hshp, unsigned int i) {
  assert(i >= 0 && i < hshp->hsh->csz); 
  return (char *)(hshp->st->ap[i]);
}

unsigned int iHshpUpcoming(hashp hshp) {    /* next number that will be assigned; if 
					       caller doesn't like it, caller can skip
					       it using SkipHshp() */
  return iHshUpcoming(hshp->hsh);
}

void SkipHshp(hashp hshp) {              /* see iHshpUpcoming() */
  SkipHsh(hshp->hsh);
  PushStack(hshp->st, NULL);   /* if someone tries to get the printname of the 
				  number we're skipping, they'll get back NULL */
}


/* -------------------------------------------------- */

/* Hash function from Dragon book, p436, by way of /pkg/libg++-2.8.0/libg++/src/hash.cc */
unsigned int iHashPJW(const char* x) 
{
  unsigned int h = 0;
  unsigned int g;

  while (*x != 0)
  {
    h = (h << 4) + *x++;
    if ((g = h & 0xF0000000) != 0)
      h = (h ^ (g >> 24)) ^ g;
  }
  return h;
}

/* slightly different version, from /pkg/binutils-2.6/src/binutils-2.6/bfd/hash.c */
unsigned long iHashElf(name)
     const unsigned char *name;
{
  unsigned long h = 0;
  unsigned long g;
  int ch;

  while ((ch = *name++) != '\0')
    {
      h = (h << 4) + ch;
      if ((g = (h & 0xf0000000)) != 0)
	{
	  h ^= g >> 24;
	  h &= ~g;
	}
    }
  return h;
}


/* Hash function from ispell, by way of /pkg/gnu/ispell-4.0/hash.c  */
unsigned int iHashIspell (p)
  register char *p;
{
  register int acc = 0;
  register unsigned char b0;

  while (1)
    {
      if ((b0 = *p++) == 0)
	break;
      acc <<= 1;
      acc += b0;
      acc += (*p << 8);
      if (acc < 0)
	{
	  acc &= (((unsigned) -1) >> 1);
	  acc++;
	}
      if (*p++ == 0)
	break;
    }

  return (acc & 0x7fff);
}


/* A randomized hash function - each character is hashed according to
   a different universal hash function depending on its position, and
   these values are combined as in iHashPJW.  !!! Caller is
   responsible for calling seed48. */
unsigned int iHashPJWRandMult(const unsigned char *x) {
  static unsigned int auMult[MAX_STR_LEN];
  static unsigned int auAdd[MAX_STR_LEN];
  static bool fInit = 0;
  unsigned int h = 0;
  unsigned int g;
  int i;

  if (fInit==0) {
    fInit=1;
    for (i=0; i< MAX_STR_LEN; i++) {
      auMult[i] = lrand48() % 257;     /* note: 257 is prime and larger than max char val */
      auAdd[i] = lrand48() % 257;
    }
  }

  for (i=0; *x != 0; i++)
  {
    assert(i < MAX_STR_LEN);
    h = (h << 4) + ((*x++) * auMult[i] + auAdd[i]) % 257;
    if ((g = h & 0xF0000000) != 0)
      h = (h ^ (g >> 24)) ^ g;
  }
  return h;
}


/* Same as iHashPJWRandMult, but moves a 2-byte window down the
   string, 1 byte at a time. 

   This function is quite effective, but slow because of the mod and
   mult.  Can replace "% 65537" with "& 0xFFFF" (equiv. to % 65536)
   which is much faster.  But this does hurt hashing performance
   (i.e., avg # probes needed at a given table size) because 65536 is
   not a prime - we lose the universal hash property in that we can
   get two double-bytes hashing to the same number.  As for the mult,
   replacing it with an xor hurts performance further.  Neither of
   these moves is at all intolerable, however - there's still a very
   clear net speedup.  */

unsigned int iHashPJWRandMult16(const unsigned char *x) {
  static unsigned int auMult[MAX_STR_LEN];
  static unsigned int auAdd[MAX_STR_LEN];
  static bool fInit = 0;
  unsigned int h = 0;
  unsigned int g;
  int i;

  if (fInit==0) {
    fInit=1;
    for (i=0; i< MAX_STR_LEN; i++) {
      auMult[i] = lrand48();     
      auAdd[i] = lrand48();
    }
  }

  for (i=0; *x != 0; i++, x++)
  {
    assert(i < MAX_STR_LEN);
    h = (h << 4) + (((*x<<8 | *(x+1)) * auMult[i] + auAdd[i]) % 65537);  /* note: 65537 is prime and larger than max 16-bit val 
									    Also: ok to say *(x+1), since that is \0 at worst */
    if ((g = h & 0xF0000000) != 0)
      h = (h ^ (g >> 24)) ^ g;
    /* if (*++x == 0) break;   !!! this line would hash faster by moving window 2 bytes at a time; hurts hashing performance, though */
  }
  return h;
}

/* Like iHashPJWRandMult16 (q.v.), but replaces the slow multiply-add-mod
   part with an xor.  Hashing function is not quite as good but much faster. */

unsigned int iHashPJWRandXor16(const unsigned char *x) {
  static unsigned int auXor[MAX_STR_LEN];
  static bool fInit = 0;
  unsigned int h = 0;
  unsigned int g;
  int i;

  if (fInit==0) {
    fInit=1;
    for (i=0; i< MAX_STR_LEN; i++) 
      auXor[i] = lrand48() & 0xFFFF;     
  }

  for (i=0; *x != 0; i++, x++)
  {
    assert(i < MAX_STR_LEN);
    h = (h << 4) + ((*x<<8 | *(x+1)) ^ auXor[i]);  /* ok to say *(x+1), since that is \0 at worst */
    if ((g = h & 0xF0000000) != 0)
      h = (h ^ (g >> 24)) ^ g;
  }
  return h;
}

    

char *strdup_nofree(char *sz, allocator all) {   /* just like strdup(), but obtains its memory 
						    without using malloc(): this is more efficient
						    in both time and space.  The tradeoff is that
						    the memory can't be freed unless the whole allocator is. 
						    We round up the allocation request to the
						    nearest multiple of sizeof(long), so that 
						    the string is properly aligned ... in case we want
						    to cast it to an array of some other type.
						    !!! The use of "long" as the most restrictive 
						    alignment type is probably not portable.  Really,
						    the alignment should be a parameter of the hash table.
						    */
  size_t siz = strlen(sz)+1;
  if (siz % sizeof(long)) 
    siz += sizeof(long) - (siz % sizeof(long));
  return strcpy((char *)pAllocateMany(all, siz), sz);
}  


