/* hsh: A stripped-down hash table that, when given a string, can return a UNIQUE small
   integer assigned to it.  (The first string ever added to the hash table is numbered 0,
   the second is numbered 1, etc.)  The caller can use these numbers to index an array of
   values of any type, which can be easily realloced or otherwise extended if it fills up.

   Optionally, the caller can also get back a canonical copy of the string -- i.e., a
   standard print name.  Such canonical copies can be used as values in a table as above.
   This lets us boil down strings to tokens and then print them out again, for instance. 

   The larger structure hshp is basically the same, but automatically maintains 
   print names as above.  In C++, it would be a subclass.  

   !!! At present, I've included no method for removing hash elements,
   since I don't currently need to do that.  It would involve freeing
   the hash table's copy of the key string (assuming I did want to get
   that memory back!).  If I were using strdup to make that copy, it
   would be easy.  But at present, I use my own allocator inside
   strdup_nofree.  So I can presently only free the whole hash table
   and its allocator at once. */

#ifndef _HASH_H
#define _HASH_H

#include "allocate.h"
#include "stack.h"

typedef struct {    /* hsh: the stripped-down hash table described above. */
                       
  char **asz;                   /* malloced array of strings */
  allocator allsz;              /* where the strings are actually stored */
  unsigned int *ai;             /* parallel malloced array of integers */
  unsigned int csz;             /* current number of entries */
  unsigned int cszMac;          /* current capacity */
  unsigned int diCollision;     /* a number in (0, cszMac) that we'll add to the bucket # in case of collision */
  unsigned int iNext;           /* next integer that will be assigned */
  unsigned int iLookups;        /* number of times we've looked something up since last resizing */
  unsigned int iCollisions;     /* number of collisions since last resizing */
  char *szName;                 /* for tracing etc. */
} *hash;

hash hshNew(unsigned int csz, char *szName);   
unsigned int iHshEntries(hash hsh);
void FreeHsh(hash hsh);
unsigned int iHashedFromSz(hash hsh, char *sz, char **psz);
int iHashedFromSzNoCreate(hash hsh, char *sz);
unsigned int *piHashedFromSz(hash hsh, char *sz, int iDefault);
unsigned int iHshUpcoming(hash hsh);
void SkipHsh(hash hsh);


/* -------------------------------------------------- */

typedef struct {    /* hshp: a version of hsh that can recover printnames. */
  hash hsh;
  stack st;     /* vector of printnames */
} *hashp;
  

hashp hshpNew(unsigned int csz, char *szName);   
void FreeHshp(hashp hshp);   
unsigned int iHashedPrintableFromSz(hashp hshp, char *sz);
int iHashedPrintableFromSzNoCreate(hashp hshp, char *sz);
char *szHshpPrintname(hashp hshp, unsigned int i);   
unsigned int iHshpEntries(hashp hshp);   /* could replace this with an iterator that calls szHshpPrintname */
unsigned int iHshpUpcoming(hashp hshp);
void SkipHshp(hashp hshp);

#endif /* _HASH_H */

